ul,
ol {
    padding-left: 18px;
}

#viz_details_main+span,
#viz_details_labels+span,
#viz_details_style+span {
    font-weight: bold;
}

.table {
    width: auto;
}

.input-container {
    position: relative;
}

.input-container .copy-icon {
    position: absolute;
    top: 5px;
    right: 10px;
    cursor: pointer;
}

/* Adjust the size and color of the icon as needed */
.input-container .copy-icon {
    font-size: 16px;
    color: #333;
}

.btn {
    overflow: hidden;
    white-space: nowrap;
    text-overflow: ellipsis;
}

.snippet th,
.snippet td {
    text-align: center;
}

table.dataTable tfoot td,
table.dataTable thead td {
    padding: 0 0 0 0;
}

td.top {
    padding-top: 33px;
}

td.top_small {
    padding-top: 5px;
}

td.top_mini {
    padding-top: 1px;
}

.well {
    padding: 7px;
    margin-bottom: 5px;
}

.col-sm-4 {
    max-width: 315px;
}

.checkbox {
    margin-top: 0;
    margin-bottom: 0;
}

.form-group {
    margin-top: 0;
    margin-bottom: 0;
}

.radio-inline+.radio-inline,
.checkbox-inline+.checkbox-inline {
    margin-left: 0;
    margin-right: 5px;
}

.radio-inline,
.checkbox-inline {
    margin-right: 5px;
}

.help-modal-dialog {
    padding-top: 80px;
}

.modal-sm {
    padding-top: 80px;
    width: 600px;
}

.help-modal-body {
    max-height: 600px;
    overflow-y: auto;
}

.modal-title {
    display: inline-block;
    line-height: 1;
    margin-top: 0;
    margin-bottom: 0;
}

img {
    max-width: 85% !important;
    height: auto;
}

/* based on https://stackoverflow.com/a/30338814/1974918*/
pre {
    overflow: auto;
    white-space: pre;
    word-wrap: normal;
    background-color: #ffffff;
}

code,
pre code {
    overflow: auto;
    white-space: pre;
    word-wrap: normal;
}

#state_download {
    font-family: "Helvetica Neue", Helvetica, Arial, sans-serif;
    font-size: 16px;
    color: black;
}

/* allow all textareas and multiple-select inputs to resize vertically */
/* have to set overflow to make resize work on Safari                   */
/* based on https://css-tricks.com/almanac/properties/r/resize/         */
textarea,
select {
    resize: vertical;
    overflow: auto;
    max-width: 100%;
}

.alignleft {
    float: left;
}

.alignright {
    float: right;
}

.aligncenter {
    margin-left: 42%;
}

.color_bar {
    background-color: blue;
}

/* Needed in combination with navbarPage with fixed-top */
/* based on https://stackoverflow.com/a/19231861/1974918 */
body {
    padding-top: 60px;
}

.navbar {
    margin-bottom: 5px;
    background-color: black;
}

.navbar .nav>li>a {
    padding-left: 7px;
    padding-right: 8px;
}

.nav-tabs {
    margin-bottom: 10px;
}

.dropdown-menu {
    max-height: 90vh;
    overflow-y: auto;
}

@media (max-width: 1000px) {
    .navbar-header {
        float: none;
    }

    .navbar-left,
    .navbar-right {
        float: none !important;
    }

    .navbar-toggle {
        display: block;
    }

    .navbar-collapse {
        border-top: 1px solid transparent;
        box-shadow: inset 0 1px 0 rgba(255, 255, 255, 0.1);
    }

    .navbar-fixed-top {
        top: 0;
        border-width: 0 0 1px;
    }

    .navbar-collapse.collapse {
        display: none !important;
    }

    .navbar-nav {
        float: none !important;
        margin-top: 7.5px;
    }

    .navbar-nav>li {
        float: none;
    }

    .navbar-nav>li>a {
        padding-top: 10px;
        padding-bottom: 10px;
    }

    .collapse.in {
        display: block !important;
    }
}

/* border radius same as other buttons in shiny */
.btn-file-solitary {
    border-radius: 4px !important;
}